.TH semaphore l "1 April 2004"
.SH NAME
semaphore - a Korn shell implementation of a counting semaphore
.SH SYNOPSIS
semaphore [other options] [user options] [directives] name
.br
semaphore [other options]
.br
semaphore -T [num ...]
.SH DESCRIPTION
semaphore is a Korn shell script that implements a counting semaphore.
Options -I, -P, and -V are the directives used to implement the Init,
P (test or try), and V (increment) functions respectively. If the -P and
-V options are used together, then the order in which the options appear
on the command line determines which operation is performed first.
.P
The -q option may be used to put processes that are waiting for a
resource in a queue. When a resource becomes available, the process at
the head of the queue gets it. Without the -q option, the order in which
waiting processes obtain a resource is unpredictable.
.P
By default, users have their own set of semaphores that are isolated
from the semaphores in other user spaces. Semaphores may be shared among
users by specifying an alternate user space using the -u option. User
spaces are implemented as directories so you must ensure that the
directory representing the user space has the appropriate permissions
using the -o and -m options as necessary.
.P
The other options may be used without specifying a semaphore
by name. If a semaphore is not specified, then the operation will be
applied to all of the semaphores in the user space. For example, the
command "semaphore -r" deletes ALL of the user's semaphores. If no
command line arguments are supplied or if an invalid option is used,
then user documentation is displayed.
.P
The implementation of semaphore hinges on the behavior of
the ln command. To implement a semaphore correctly, the test of a
semaphore's value and the decrementing of this value must be an atomic
operation. In the function grab_resource, the act of creating a symbolic
link does both things at once. If the ln command is successful, then
the semaphore's count is effectively tested and decremented in one atomic
step.
.SS BACKGROUND
Semaphores, as described by Dijkstra, are the classic method for
restricting access to shared resources in a multi-processing environment.
.P
A counting semaphore maintains a count of the available resources for
a particular semaphore. A counting semaphore with only one resource
is called a binary semaphore.
.P
The functions Init, P, and V are used to manipulate a semaphore.
The letter P stands for the Dutch word proberen, meaning "test"
or "try", and V stands for verhogen, meaning "increment".
.P
Init initializes a semaphore by specifying how many resources the
semaphore has.
.P
P tries to obtain a resource. It tests to see if the semaphore's value
is greater than zero. If it is, then the function decrements the value
of the semaphore. If no resources are available, then the function waits
for one to be freed by the V function.
.P
V makes a resource available by incrementing the value of the semaphore.
.SH OPTIONS
.SS PRIMARY DIRECTIVES
.TP
-I num
Initialize the semaphore and set the number of resources.
.TP
-P
Grab a resource and decrement the value of the semaphore. Wait until a
resource become available if necessary.
.TP
-V
Increment the value of the semaphore if and only if the calling process has
a resource to return.
.SS OTHER OPTIONS
.TP
-d
Run semaphore in debug mode. Additional error messages may
be displayed. Function tracing is turned on.
.TP
-l
List information about the semaphore(s).
.TP
-m mode
Set the permissions mode on the directory used to implement the
user space.
.TP
-o owner[:group]
Set the ownership on the directory used to implement the
user space.
.TP
-p pid
Use with -V to explicity specify the pid associated with a resource to
be freed. Although not recommended, -p may also be used with -P to
explicity specify the pid associated with a resource.
.TP
-q
Use with -P to place processes that are waiting for a resource
in a queue.
.TP
-r
Remove the specified semaphore, or remove all semaphores in
the user space if no operand is supplied. If -I and -r are
used together, then the semaphore is first removed and then
initialized.
.TP
-s secs
Set the number of seconds the P operation will sleep between
tests. By default, processes waiting for a resource will
sleep for one second before trying again.
.TP
-t tries
Set the number of times the P operation will try to obtain a
resource before timing out if none are available. By default,
the P operation will keep trying until is a resource is
obtained.
.TP
-u user
Specify an alternate user space for the semaphore.
.TP
-v
Execute semaphore in verbose mode.
.TP
-w
Wipe semaphore. Wiping returns any resources that are being
held by processes that are no longer running. If -w and -l
are used together, semphores are wiped first and then listed.
Semaphores are automatically wiped before the P operation is
performed, so you do not have to use the -w option with the -P
directive to ensure that the semaphore is wiped.
.SS TESTING OPTIONS
.TP
-T
If no operands are supplied than semaphore displays
a one-line message indicating the number of test cases; otherwise,
semaphore executes the specified tests. Test zero (0) is a special
test case that executes all of the other test cases in sequential
order.
.SH OPERANDS
.TP
name
Identifies a semaphore by name.
.TP
num
When the -T option is used, num is a non-negative integer that
references a test case to be executed.
.SH EXAMPLES
System administrators often encounter situations where they don't want more
than one person or process to be doing something at the same time.
For example, if you schedule a backup job using cron, you probably
don't want another instance of the job starting if the previous instance
hasn't finished yet. Or you may have developed a script to add user accounts
to the system, and you want to prevent different administrators from running
the script at the same time. Or maybe you need to ensure that you don't
exceed the limit for the maximum number of concurrent users who are licensed
to use a software package.
.P
To solve the problem where only one resource is available, you could use a
binary semaphore as follows:
.P
> #!/bin/ksh
.br
>
.br
> if semaphore -P MY_PROGRAM
.br
> then
.br
>     my_program
.br
>     semaphore -V MY_PROGRAM
.br
> else
.br
>     echo "Unexpected error!"
.br
> fi
.P
Or you may decided to take a slightly different approach:
.P
> #!/bin/ksh
.br
>
.br
> if semaphore -t 10 -s 5 -P MY_PROGRAM
.br
> then
.br
>     my_program
.br
>     semaphore -V MY_PROGRAM
.br
> else
.br
>     echo "Sorry, my_program is currently in use."
.br
>     echo "Stopped waiting after 10 tries and 50 seconds."
.br
> fi
.P
In the case where there is more than one resource to be shared, you could use a counting semaphore as follows:
.P
> #!/bin/ksh
.br
>
.br
> semaphore -I 5 OUR_SOFTWARE
.br
>
.br
> if  semaphore -P OUR_SOFTWARE
.br
> then
.br
>     our_software
.br
>     semaphore -V OUR_SOFTWARE
.br
> else
.br
>     echo "There are 5 instances of our_software in use."
.br
>     echo "Please try again later."
.br
> fi
.P
A system administrator who is not familiar with
semaphores might create a flag file or run ps and grep for a process id to
see if a program is currently running. A problem with the flag file approach
is that if the process dies unexpectedly and doesn't clean up after itself,
then the file could hang around indefinitely, making other processes think
the program is still running when it isn't. Even worse, both the flag file
and the ps approach suffer from a potential race condition. For example,
a script might attempts to emulate a semaphore as follows:
.P
> #!/bin/ksh
.br
>
.br
> while (( $(ps | grep my_program | grep -c -v grep) > 0 ))
.br
> do
.br
>     sleep 1
.br
> done
.br
>
.br
> my_program
.P
However, in the example above two instances of the script could execute ps at
almost the same time. If both calls to ps completed before either process
could start my_program, then you could end up with two instances of
my_program running at the same time. Semaphores guard against this type of
race condition.
.SH EXIT STATUS
The following exit values are returned:
.TP
0
The command completed successfully.
.TP
>0
An error occurred or a resource was not obtained.
.SH FILES
.TP
/tmp/semaphores/
The ROOT directory where all semaphore files are stored. To use a different
directory, change the value of ROOT_DIR in .../semaphore/bin/semaphore.
.TP
ROOT/user/
The directory containing all semaphore files for a particular user space.
.TP
ROOT/user/semaphore/
The directory containing files for a particular semaphore.
.TP
ROOT/user/semaphore/resources
A regular file that contains an integer representing the number of resources
a semaphore has.
.TP
ROOT/user/semaphore/n -> pid
A symbolic link signifying that resource n for a particular semaphore is
currently taken by process pid.
.TP
ROOT/user/semaphore/queue
A directory containing files used to implement a queue for processes that
are waiting for a resource to become available.
.SH SEE ALSO
semaphore(3T), semaphore(9F)
.SH BUGS
Since process ids are recyled, it is possible that a process could grab
a resource, die without returning it, and another process could start
running with the same process id. In that case, the P function could
mistakenly report that no resources are available, when in fact a
resource really is available. The probability of this happening is
small, and the impact - if it were to happen  - is relatively minor,
since the resource would eventually be freed when the process with
the recycled process id terminates.
.P
The semaphore directory should not be located on a file system that is
shared between two hosts (via NFS, for example). Otherwise, the
wipe_semaphore function could mistakenly conclude that a process is no
longer running (when in fact it IS running on another host) and free a
resource that is not ready to be freed yet.
.SH REFERENCES
The Free On-line Dictionary of Computing, available at
http://www.foldoc.org/
.br
Advanced Programming in the UNIX Environment, by W. Richard Stevens
.SH VERSION
Version 0.1 (beta) - released on 11 MAR 2004
.br
Version 0.2 (beta) - released on 1 APR 2004
.br
Version 0.3 (beta) - WIP
.SH AUTHOR
John Spurgeon
.br
Intel Corporation
.br
5200 NE Elam Young Parkway
.br
Hillsboro, OR 97124
.P
john.p.spurgeon@intel.com
